"""The graphql plugin of the Minos Framework."""

__author__ = "Minos Framework Devs"
__email__ = "hey@minos.run"
__version__ = "0.8.0.dev1"

from .builders import (
    GraphQLSchemaBuilder,
)
from .decorators import (
    GraphQlCommandEnrouteDecorator,
    GraphQlEnroute,
    GraphQlEnrouteDecorator,
    GraphQlQueryEnrouteDecorator,
)
from .handlers import (
    GraphQlHandler,
)
from .routers import (
    GraphQlHttpRouter,
)


def _register_enroute():
    GraphQlEnroute.register()



from abc import (
    ABC,
)
from collections.abc import (
    Iterable,
)
from typing import (
    Final,
    Optional,
)

from minos.networks import (
    EnrouteDecorator,
    EnrouteDecoratorKind,
    enroute,
)


class GraphQlEnrouteDecorator(EnrouteDecorator, ABC):
    """GraphQl Enroute Decorator class"""

    def __init__(self, name: str, output, argument: Optional = None, **kwargs):
        super().__init__(**kwargs)
        self.name = name
        self.argument = argument
        self.output = output

    def __iter__(self) -> Iterable:
        yield from (
            type(self),
            self.name,
        )


class GraphQlCommandEnrouteDecorator(GraphQlEnrouteDecorator):
    """GraphQl Command Enroute Decorator class"""

    KIND: Final[EnrouteDecoratorKind] = EnrouteDecoratorKind.Command


class GraphQlQueryEnrouteDecorator(GraphQlEnrouteDecorator):
    """GraphQl Query Enroute Decorator class"""

    KIND: Final[EnrouteDecoratorKind] = EnrouteDecoratorKind.Query


class GraphQlEnroute:
    """GraphQl Enroute class"""

    command = GraphQlCommandEnrouteDecorator
    query = GraphQlQueryEnrouteDecorator

    @classmethod
    def register(cls):
        """Register the graphql sub-enroute."""
        # noinspection PyProtectedMember
        enroute._register_sub_enroute("graphql", cls)



"""This module contains the implementation of the graphql handler."""

import logging
import traceback
from typing import (
    Any,
)

from graphql import (
    ExecutionResult,
    GraphQLError,
    GraphQLSchema,
    graphql,
    print_schema,
)

from minos.networks import (
    Request,
    Response,
    ResponseException,
)

logger = logging.getLogger(__name__)


class GraphQlHandler:
    """GraphQl Handler"""

    def __init__(self, schema: GraphQLSchema):
        self._schema = schema

    async def execute_operation(self, request: Request) -> Response:
        """Execute incoming request extracting variables and passing to graphql.

        :param request: The request containing the graphql operation.
        :return: A response containing the graphql result.
        """
        arguments = await self._build_graphql_arguments(request)
        result = await graphql(schema=self._schema, **arguments)

        return self._build_response_from_graphql(result)

    @staticmethod
    async def _build_graphql_arguments(request: Request) -> dict[str, Any]:
        content = await request.content()
        source = dict()
        variables = dict()

        if isinstance(content, str):
            source = content
        elif isinstance(content, dict):
            if "query" in content:
                source = content["query"]
            if "variables" in content:
                variables = content["variables"]

        return {"source": source, "variable_values": variables}

    def _build_response_from_graphql(self, result: ExecutionResult) -> Response:
        content = {"data": result.data}
        if result.errors is not None:
            content["errors"] = [err.message for err in result.errors]
            self._log_errors(result.errors)

        status = self._get_status(result)

        return Response(content, status=status)

    @staticmethod
    def _get_status(result: ExecutionResult) -> int:
        status = 200
        for error in result.errors or []:
            if error.original_error is None:
                current = 400
            elif isinstance(error.original_error, ResponseException):
                current = error.original_error.status
            else:
                current = 500
            status = max(status, current)
        return status

    @staticmethod
    def _log_errors(errors: list[GraphQLError]) -> None:
        for error in errors:
            if error.original_error is None:
                tb = repr(error)
            else:
                tb = "".join(traceback.format_tb(error.__traceback__))

            if error.original_error is None or isinstance(error.original_error, ResponseException):
                logger.error(f"Raised an application exception:\n {tb}")
            else:
                logger.exception(f"Raised a system exception:\n {tb}")

    async def get_schema(self, request: Request) -> Response:
        """Get schema

        :param request: An empty request.
        :return: A Response containing the schema as a string.
        """
        return Response(print_schema(self._schema))



from __future__ import (
    annotations,
)

from collections.abc import (
    Callable,
)
from functools import (
    wraps,
)
from inspect import (
    isawaitable,
)
from typing import (
    Any,
    Awaitable,
    Optional,
    Union,
)

from graphql import (
    GraphQLArgument,
    GraphQLField,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLString,
)

from minos.networks import (
    EnrouteDecoratorKind,
    InMemoryRequest,
    Request,
    Response,
)

from ..decorators import (
    GraphQlEnrouteDecorator,
)


class GraphQLSchemaBuilder:
    """GraphQL Schema Builder class."""

    def __init__(self, *args, **kwargs):
        self.schema = GraphQLSchema(**kwargs)

    @classmethod
    def build(cls, routes: dict[GraphQlEnrouteDecorator, Callable]) -> GraphQLSchema:
        """Build a new schema from routes.

        :param routes: The routes to build the schema.
        :return: A ``GraphQLSchema`` instance.
        """
        schema_args = cls._build(routes)
        return cls(**schema_args).schema

    @classmethod
    def _build(cls, routes: dict[GraphQlEnrouteDecorator, Callable]) -> dict[str, Optional[GraphQLObjectType]]:
        query = cls._build_queries(routes)
        mutation = cls._build_mutations(routes)

        return {"query": query, "mutation": mutation}

    @staticmethod
    def adapt_callback(
        callback: Callable[[Request], Union[Optional[Response], Awaitable[Optional[Response]]]]
    ) -> Callable[[Any, Any, Any], Awaitable[Any]]:
        """Adapt a callback from framework's Request-Response to GraphQl structure.

        :param callback: The callback to be adapted.
        :return: The adapted callback.
        """

        @wraps(callback)
        async def _wrapper(_source, _info, request: Any = None):
            request = InMemoryRequest(request)

            response = callback(request)
            if isawaitable(response):
                response = await response

            return await response.content()

        return _wrapper

    @classmethod
    def _build_queries(cls, routes: dict[GraphQlEnrouteDecorator, Callable]) -> GraphQLObjectType:
        fields = dict()
        for route, callback in routes.items():
            callback = cls.adapt_callback(callback)
            if route.KIND == EnrouteDecoratorKind.Query:
                fields[route.name] = cls._build_field(route, callback)

        if not len(fields):
            fields["dummy"] = GraphQLField(
                type_=GraphQLString,
                description="Dummy query added to surpass the 'Type Query must define at least one field' constraint.",
            )

        return GraphQLObjectType("Query", fields=fields)

    @classmethod
    def _build_mutations(cls, routes: dict[GraphQlEnrouteDecorator, Callable]) -> Optional[GraphQLObjectType]:
        fields = dict()
        for route, callback in routes.items():
            callback = cls.adapt_callback(callback)

            if route.KIND == EnrouteDecoratorKind.Command:
                fields[route.name] = cls._build_field(route, callback)

        if not len(fields):
            return None

        return GraphQLObjectType("Mutation", fields=fields)

    @staticmethod
    def _build_field(item: GraphQlEnrouteDecorator, callback: Callable) -> GraphQLField:
        argument = item.argument
        output = item.output

        args = None
        if argument is not None:
            args = {"request": GraphQLArgument(argument)}

        return GraphQLField(output, args=args, resolve=callback)



from __future__ import (
    annotations,
)

from typing import (
    Callable,
)

from minos.networks import (
    EnrouteDecorator,
    HttpEnrouteDecorator,
    HttpRouter,
)

from .builders import (
    GraphQLSchemaBuilder,
)
from .decorators import (
    GraphQlEnrouteDecorator,
)
from .handlers import (
    GraphQlHandler,
)


class GraphQlHttpRouter(HttpRouter):
    """GraphQl Http Router class."""

    def _filter_routes(self, routes: dict[EnrouteDecorator, Callable]) -> dict[EnrouteDecorator, Callable]:
        routes = {
            decorator: callback
            for decorator, callback in routes.items()
            if isinstance(decorator, GraphQlEnrouteDecorator)
        }
        schema = GraphQLSchemaBuilder.build(routes)
        handler = GraphQlHandler(schema)
        service_name = self._config.get_name().lower()
        return {
            HttpEnrouteDecorator(f"/{service_name}/graphql", "POST"): handler.execute_operation,
            HttpEnrouteDecorator(f"/{service_name}/graphql/schema", "GET"): handler.get_schema,
        }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['minos', 'minos.plugins.graphql', 'minos.plugins.graphql.builders']

package_data = \
{'': ['*']}

install_requires = \
['graphql-core>=3.2.0,<4.0.0',
 'minos-microservice-common',
 'minos-microservice-networks']

setup_kwargs = {
    'name': 'minos-router-graphql',
    'version': '0.8.0.dev1',
    'description': 'The graphql plugin of the Minos Framework',
    'long_description': '<p align="center">\n  <a href="https://minos.run" target="_blank"><img src="https://raw.githubusercontent.com/minos-framework/.github/main/images/logo.png" alt="Minos logo"></a>\n</p>\n\n## minos-router-graphql\n\n[![PyPI Latest Release](https://img.shields.io/pypi/v/minos-broker-kafka.svg)](https://pypi.org/project/minos-broker-kafka/)\n[![GitHub Workflow Status](https://img.shields.io/github/workflow/status/minos-framework/minos-python/pages%20build%20and%20deployment?label=docs)](https://minos-framework.github.io/minos-python)\n[![License](https://img.shields.io/github/license/minos-framework/minos-python.svg)](https://github.com/minos-framework/minos-python/blob/main/LICENSE)\n[![Coverage](https://codecov.io/github/minos-framework/minos-python/coverage.svg?branch=main)](https://codecov.io/gh/minos-framework/minos-python)\n[![Stack Overflow](https://img.shields.io/badge/Stack%20Overflow-Ask%20a%20question-green)](https://stackoverflow.com/questions/tagged/minos)\n\n## Summary\n\nThis is graphQL plugin for Minos framework. This plugin integrates the official [graphql-core](https://github.com/graphql-python/graphql-core) library. It is oriented to facilitate the development and better organize the graphql code.\n\n## Installation\n\nInstall the dependency:\n\n```shell\npip install minos-router-graphql\n```\n\nModify `config.yml` file:\n\n```yaml\n...\nrouters:\n  - minos.plugins.graphql.GraphQlRouter\n...\n```\n\n## How to use\n\n### Define your business operation\n\nWe will use simple query for this demonstration:\n\n```python\nfrom graphql import (\n    GraphQLString,\n)\n\nfrom minos.networks import (\n    Request,\n    Response,\n    enroute,\n)\n\n\nclass QueryService:\n    @enroute.graphql.query(name="SimpleQuery", output=GraphQLString)\n    def simple_query(self, request: Request):\n        ...\n\n        return Response("ABCD")\n```\n\n### Execute query\n\nSend `post` request to `http://your_ip_address:port/service_name/graphql` endpoint:\n\n```json\n{\n  "query": "{ SimpleQuery }"\n}\n```\n\nYou will receive:\n\n```json\n{\n  "data": {\n    "SimpleQuery": "ABCD"\n  },\n  "errors": []\n}\n```\n\nThat\'s all you need to make it work !\n\nFor more information about graphql and how to define fields or structures, please see the official [graphql-core](https://github.com/graphql-python/graphql-core). library.\n\n## Decorators\n\nThere are 2 types of decorators, one for `queries` and one for `mutations` (commands).\n\n```python\n@enroute.graphql.query(name="TestQuery", argument=GraphQLString, output=GraphQLString)\ndef test_query(self, request: Request):\n    ...\n    return Responnse(...)\n\n\n@enroute.graphql.command(name="TestCommand", argument=GraphQLString, output=GraphQLString)\ndef test_command(self, request: Request):\n    ...\n    return Responnse(...)\n```\n\nBoth decorators have the following arguments:\n\n- `name`: The name of the query or command\n- `argument` [Optional]: The arguments it receives, if any.\n- `output`: The expected output.\n\n### Resolvers\n\nAs you have seen above, the decorator does not specify the function that will resolve it.\n\nThis is because it automatically takes the decorated function.\n\nIn the following example:\n\n```python\n@enroute.graphql.query(name="TestQuery", argument=GraphQLString, output=GraphQLString)\ndef test_query(self, request: Request):\n    ...\n    return Responnse(...)\n```\n\nThe function in charge of resolving the query is the decorated function `test_query`.\n\n### Queries (Query Service)\n\nQueries are used for a single purpose as their name indicates and that is to obtain information, that is, for queries.\n\nBase structure example:\n\n```python\nclass QueryService:\n    @enroute.graphql.query(name="TestQuery", argument=GraphQLString, output=GraphQLString)\n    def test_query(self, request: Request):\n        ...\n        return Responnse(...)\n```\n\nMore complex example:\n\n```python\nfrom graphql import (\n    GraphQLBoolean,\n    GraphQLField,\n    GraphQLID,\n    GraphQLInt,\n    GraphQLNonNull,\n    GraphQLObjectType,\n    GraphQLString,\n)\nfrom typing import (\n    NamedTuple,\n    Optional,\n)\nfrom minos.networks import (\n    Request,\n    Response,\n    enroute,\n)\n\nuser_type = GraphQLObjectType(\n    "UserType",\n    {\n        "id": GraphQLField(GraphQLNonNull(GraphQLID)),\n        "firstName": GraphQLField(GraphQLNonNull(GraphQLString)),\n        "lastName": GraphQLField(GraphQLNonNull(GraphQLString)),\n        "tweets": GraphQLField(GraphQLInt),\n        "verified": GraphQLField(GraphQLNonNull(GraphQLBoolean)),\n    },\n)\n\n\nclass User(NamedTuple):\n    """A simple user object class."""\n\n    firstName: str\n    lastName: str\n    tweets: Optional[int]\n    id: Optional[str] = None\n    verified: bool = False\n\n\nclass QueryService:\n    @enroute.graphql.query(name="GetUser", argument=GraphQLInt, output=user_type)\n    def test_query(self, request: Request):\n        id = await request.content()\n        return Response(User(firstName="Jack", lastName="Johnson", tweets=563, id=str(id), verified=True))\n```\n\nIf you POST `{service_name}/graphql` endpoint passing the query and variables:\n\n```json\n{\n  "query": "query ($userId: Int!) { GetUser(request: $userId) {id firstName lastName tweets verified}}",\n  "variables": {\n    "userId": 3\n  }\n}\n```\n\nYoy will receive:\n\n```json\n{\n  "data": {\n    "GetUser": {\n      "id": "3",\n      "firstName": "Jack",\n      "lastName": "Johnson",\n      "tweets": 563,\n      "verified": true\n    }\n  },\n  "errors": []\n}\n```\n\n### Mutations (Command Service)\n\nMutations are used to create, update or delete information.\n\nBase structure example:\n\n```python\nclass CommandService:\n    @enroute.graphql.command(name="TestQuery", argument=GraphQLString, output=GraphQLString)\n    def test_command(self, request: Request):\n        ...\n        return Responnse(...)\n```\n\nMore complex example:\n\n```python\nfrom graphql import (\n    GraphQLBoolean,\n    GraphQLField,\n    GraphQLID,\n    GraphQLInputField,\n    GraphQLInputObjectType,\n    GraphQLInt,\n    GraphQLNonNull,\n    GraphQLObjectType,\n    GraphQLString,\n)\nfrom typing import (\n    NamedTuple,\n    Optional,\n)\nfrom minos.networks import (\n    Request,\n    Response,\n    enroute,\n)\n\nuser_type = GraphQLObjectType(\n    "UserType",\n    {\n        "id": GraphQLField(GraphQLNonNull(GraphQLID)),\n        "firstName": GraphQLField(GraphQLNonNull(GraphQLString)),\n        "lastName": GraphQLField(GraphQLNonNull(GraphQLString)),\n        "tweets": GraphQLField(GraphQLInt),\n        "verified": GraphQLField(GraphQLNonNull(GraphQLBoolean)),\n    },\n)\n\nuser_input_type = GraphQLInputObjectType(\n    "UserInputType",\n    {\n        "firstName": GraphQLInputField(GraphQLNonNull(GraphQLString)),\n        "lastName": GraphQLInputField(GraphQLNonNull(GraphQLString)),\n        "tweets": GraphQLInputField(GraphQLInt),\n        "verified": GraphQLInputField(GraphQLBoolean),\n    },\n)\n\n\nclass User(NamedTuple):\n    """A simple user object class."""\n\n    firstName: str\n    lastName: str\n    tweets: Optional[int]\n    id: Optional[str] = None\n    verified: bool = False\n\n\nclass CommandService:\n    @enroute.graphql.command(name="CreateUser", argument=GraphQLNonNull(user_input_type), output=user_type)\n    def test_command(self, request: Request):\n        params = await request.content()\n        return Response(\n            User(\n                firstName=params["firstName"],\n                lastName=params["lastName"],\n                tweets=params["tweets"],\n                id="4kjjj43-l23k4l3-325kgaa2",\n                verified=params["verified"],\n            )\n        )\n```\n\nIf you POST `{service_name}/graphql` endpoint passing the query and variables:\n\n```json\n{\n  "query": "mutation ($userData: UserInputType!) { CreateUser(request: $userData) {id, firstName, lastName, tweets, verified}}",\n  "variables": {\n    "userData": {\n      "firstName": "John",\n      "lastName": "Doe",\n      "tweets": 42,\n      "verified": true\n    }\n  }\n}\n```\n\nYoy will receive:\n\n```json\n{\n  "data": {\n    "CreateUser": {\n      "id": "4kjjj43-l23k4l3-325kgaa2",\n      "firstName": "John",\n      "lastName": "Doe",\n      "tweets": 42,\n      "verified": true\n    }\n  },\n  "errors": []\n}\n```\n\n### Get Schema\n\nBy calling `{service_name}/graphql/schema` with `GET` method, you will receive the schema:\n\n```text\n"type Query {\\n  GetUser(request: Int): UserType\\n}\\n\\ntype UserType {\\n  id: ID!\\n  firstName: String!\\n  lastName: String!\\n  tweets: Int\\n  verified: Boolean!\\n}\\n\\ntype Mutation {\\n  CreateUser(request: UserInputType!): UserType\\n}\\n\\ninput UserInputType {\\n  firstName: String!\\n  lastName: String!\\n  tweets: Int\\n  verified: Boolean\\n}"\n```\n\n## Documentation\n\nThe official API Reference is publicly available at the [GitHub Pages](https://minos-framework.github.io/minos-python).\n\n## Source Code\n\nThe source code of this project is hosted at the [GitHub Repository](https://github.com/minos-framework/minos-python).\n\n## Getting Help\n\nFor usage questions, the best place to go to is [StackOverflow](https://stackoverflow.com/questions/tagged/minos).\n\n## Discussion and Development\n\nMost development discussions take place over the [GitHub Issues](https://github.com/minos-framework/minos-python/issues). In addition, a [Gitter channel](https://gitter.im/minos-framework/community) is available for development-related questions.\n\n## License\n\nThis project is distributed under the [MIT](https://raw.githubusercontent.com/minos-framework/minos-python/main/LICENSE) license.\n',
    'author': 'Minos Framework Devs',
    'author_email': 'hey@minos.run',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.minos.run/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
